@@ -8,27 +8,30 @@ import os
 import sys
 import traceback
 import warnings
+from collections.abc import Coroutine
 
 from multidict import CIMultiDict, MultiDict, MultiDictProxy, istr
 from yarl import URL
 
+from . import client_exceptions, client_reqrep
 from . import connector as connector_mod
-from . import client_exceptions, client_reqrep, hdrs, http, payload
+from . import hdrs, http, payload
 from .client_exceptions import *  # noqa
 from .client_exceptions import (ClientError, ClientOSError, InvalidURL,
                                 ServerTimeoutError, WSServerHandshakeError)
 from .client_reqrep import *  # noqa
-from .client_reqrep import ClientRequest, ClientResponse
+from .client_reqrep import ClientRequest, ClientResponse, _merge_ssl_params
 from .client_ws import ClientWebSocketResponse
 from .connector import *  # noqa
 from .connector import TCPConnector
 from .cookiejar import CookieJar
-from .helpers import (PY_35, CeilTimeout, TimeoutHandle, _BaseCoroMixin,
-                      deprecated_noop, proxies_from_env, sentinel,
-                      strip_auth_from_url)
+from .helpers import (DEBUG, PY_36, CeilTimeout, TimeoutHandle,
+                      proxies_from_env, sentinel, strip_auth_from_url)
 from .http import WS_KEY, WebSocketReader, WebSocketWriter
 from .http_websocket import WSHandshakeError, ws_ext_gen, ws_ext_parse
 from .streams import FlowControlDataQueue
+from .tcp_helpers import tcp_cork, tcp_nodelay
+from .tracing import Trace
 
 
 __all__ = (client_exceptions.__all__ +  # noqa
@@ -44,6 +47,16 @@ DEFAULT_TIMEOUT = 5 * 60
 class ClientSession:
     """First-class interface for making HTTP requests."""
 
+    ATTRS = frozenset([
+        '_source_traceback', '_connector',
+        'requote_redirect_url', '_loop', '_cookie_jar',
+        '_connector_owner', '_default_auth',
+        '_version', '_json_serialize', '_read_timeout',
+        '_conn_timeout', '_raise_for_status', '_auto_decompress',
+        '_trust_env', '_default_headers', '_skip_auto_headers',
+        '_request_class', '_response_class',
+        '_ws_response_class', '_trace_configs'])
+
     _source_traceback = None
     _connector = None
 
@@ -57,7 +70,8 @@ class ClientSession:
                  version=http.HttpVersion11,
                  cookie_jar=None, connector_owner=True, raise_for_status=False,
                  read_timeout=sentinel, conn_timeout=None,
-                 auto_decompress=True, trust_env=False):
+                 auto_decompress=True, trust_env=False,
+                 trace_configs=None):
 
         implicit_loop = False
         if loop is None:
@@ -81,7 +95,7 @@ class ClientSession:
 
         if implicit_loop and not loop.is_running():
             warnings.warn("Creating a client session outside of coroutine is "
-                          "a very dangerous idea", ResourceWarning,
+                          "a very dangerous idea",
                           stacklevel=2)
             context = {'client_session': self,
                        'message': 'Creating a client session outside '
@@ -96,6 +110,7 @@ class ClientSession:
 
         if cookies is not None:
             self._cookie_jar.update_cookies(cookies)
+
         self._connector = connector
         self._connector_owner = connector_owner
         self._default_auth = auth
@@ -124,10 +139,34 @@ class ClientSession:
         self._response_class = response_class
         self._ws_response_class = ws_response_class
 
+        self._trace_configs = trace_configs or []
+        for trace_config in self._trace_configs:
+            trace_config.freeze()
+
+    def __init_subclass__(cls):
+        warnings.warn("Inheritance class {} from ClientSession "
+                      "is discouraged".format(cls.__name__),
+                      DeprecationWarning,
+                      stacklevel=2)
+
+    if DEBUG:
+        def __setattr__(self, name, val):
+            if name not in self.ATTRS:
+                warnings.warn("Setting custom ClientSession.{} attribute "
+                              "is discouraged".format(name),
+                              DeprecationWarning,
+                              stacklevel=2)
+            super().__setattr__(name, val)
+
     def __del__(self, _warnings=warnings):
         if not self.closed:
+            if PY_36:
+                kwargs = {'source': self}
+            else:
+                kwargs = {}
             _warnings.warn("Unclosed client session {!r}".format(self),
-                           ResourceWarning)
+                           ResourceWarning,
+                           **kwargs)
             context = {'client_session': self,
                        'message': 'Unclosed client session'}
             if self._source_traceback is not None:
@@ -138,42 +177,38 @@ class ClientSession:
         """Perform HTTP request."""
         return _RequestContextManager(self._request(method, url, **kwargs))
 
-    @asyncio.coroutine
-    def _request(self, method, url, *,
-                 params=None,
-                 data=None,
-                 json=None,
-                 headers=None,
-                 skip_auto_headers=None,
-                 auth=None,
-                 allow_redirects=True,
-                 max_redirects=10,
-                 encoding=None,
-                 compress=None,
-                 chunked=None,
-                 expect100=False,
-                 read_until_eof=True,
-                 proxy=None,
-                 proxy_auth=None,
-                 timeout=sentinel,
-                 verify_ssl=None,
-                 fingerprint=None,
-                 ssl_context=None,
-                 proxy_headers=None):
+    async def _request(self, method, url, *,
+                       params=None,
+                       data=None,
+                       json=None,
+                       headers=None,
+                       skip_auto_headers=None,
+                       auth=None,
+                       allow_redirects=True,
+                       max_redirects=10,
+                       compress=None,
+                       chunked=None,
+                       expect100=False,
+                       read_until_eof=True,
+                       proxy=None,
+                       proxy_auth=None,
+                       timeout=sentinel,
+                       verify_ssl=None,
+                       fingerprint=None,
+                       ssl_context=None,
+                       ssl=None,
+                       proxy_headers=None,
+                       trace_request_ctx=None):
 
         # NOTE: timeout clamps existing connect and read timeouts.  We cannot
         # set the default to None because we need to detect if the user wants
         # to use the existing timeouts by setting timeout to None.
 
-        if encoding is not None:
-            warnings.warn(
-                "encoding parameter is not supported, "
-                "please use FormData(charset='utf-8') instead",
-                DeprecationWarning)
-
         if self.closed:
             raise RuntimeError('Session is closed')
 
+        ssl = _merge_ssl_params(ssl, verify_ssl, ssl_context, fingerprint)
+
         if data is not None and json is not None:
             raise ValueError(
                 'data and json parameters can not be used at the same time')
@@ -206,7 +241,7 @@ class ClientSession:
             try:
                 proxy = URL(proxy)
             except ValueError:
-                raise InvalidURL(url)
+                raise InvalidURL(proxy)
 
         # timeout is cumulative for all request operations
         # (request, redirects, responses, data consuming)
@@ -215,7 +250,23 @@ class ClientSession:
             timeout if timeout is not sentinel else self._read_timeout)
         handle = tm.start()
 
-        url = URL(url)
+        traces = [
+            Trace(
+                self,
+                trace_config,
+                trace_config.trace_config_ctx(
+                    trace_request_ctx=trace_request_ctx)
+            )
+            for trace_config in self._trace_configs
+        ]
+
+        for trace in traces:
+            await trace.send_request_start(
+                method,
+                url,
+                headers
+            )
+
         timer = tm.timer()
         try:
             with timer:
@@ -246,7 +297,8 @@ class ClientSession:
                     elif self._trust_env:
                         for scheme, proxy_info in proxies_from_env().items():
                             if scheme == url.scheme:
-                                proxy, proxy_auth = proxy_info
+                                proxy = proxy_info.proxy
+                                proxy_auth = proxy_info.proxy_auth
                                 break
 
                     req = self._request_class(
@@ -258,25 +310,31 @@ class ClientSession:
                         response_class=self._response_class,
                         proxy=proxy, proxy_auth=proxy_auth, timer=timer,
                         session=self, auto_decompress=self._auto_decompress,
-                        verify_ssl=verify_ssl, fingerprint=fingerprint,
-                        ssl_context=ssl_context, proxy_headers=proxy_headers)
+                        ssl=ssl, proxy_headers=proxy_headers, traces=traces)
 
                     # connection timeout
                     try:
                         with CeilTimeout(self._conn_timeout, loop=self._loop):
-                            conn = yield from self._connector.connect(req)
+                            conn = await self._connector.connect(
+                                req,
+                                traces=traces
+                            )
                     except asyncio.TimeoutError as exc:
                         raise ServerTimeoutError(
                             'Connection timeout '
                             'to host {0}'.format(url)) from exc
 
-                    conn.writer.set_tcp_nodelay(True)
+                    tcp_nodelay(conn.transport, True)
+                    tcp_cork(conn.transport, False)
                     try:
-                        resp = req.send(conn)
                         try:
-                            yield from resp.start(conn, read_until_eof)
-                        except:
-                            resp.close()
+                            resp = await req.send(conn)
+                            try:
+                                await resp.start(conn, read_until_eof)
+                            except BaseException:
+                                resp.close()
+                                raise
+                        except BaseException:
                             conn.close()
                             raise
                     except ClientError:
@@ -289,6 +347,15 @@ class ClientSession:
                     # redirects
                     if resp.status in (
                             301, 302, 303, 307, 308) and allow_redirects:
+
+                        for trace in traces:
+                            await trace.send_request_redirect(
+                                method,
+                                url,
+                                headers,
+                                resp
+                            )
+
                         redirects += 1
                         history.append(resp)
                         if max_redirects and redirects >= max_redirects:
@@ -352,15 +419,30 @@ class ClientSession:
                     handle.cancel()
 
             resp._history = tuple(history)
+
+            for trace in traces:
+                await trace.send_request_end(
+                    method,
+                    url,
+                    headers,
+                    resp
+                )
             return resp
 
-        except:
+        except BaseException as e:
             # cleanup timer
             tm.close()
             if handle:
                 handle.cancel()
                 handle = None
 
+            for trace in traces:
+                await trace.send_request_exception(
+                    method,
+                    url,
+                    headers,
+                    e
+                )
             raise
 
     def ws_connect(self, url, *,
@@ -375,6 +457,7 @@ class ClientSession:
                    headers=None,
                    proxy=None,
                    proxy_auth=None,
+                   ssl=None,
                    verify_ssl=None,
                    fingerprint=None,
                    ssl_context=None,
@@ -394,30 +477,31 @@ class ClientSession:
                              headers=headers,
                              proxy=proxy,
                              proxy_auth=proxy_auth,
+                             ssl=ssl,
                              verify_ssl=verify_ssl,
                              fingerprint=fingerprint,
                              ssl_context=ssl_context,
                              proxy_headers=proxy_headers,
                              compress=compress))
 
-    @asyncio.coroutine
-    def _ws_connect(self, url, *,
-                    protocols=(),
-                    timeout=10.0,
-                    receive_timeout=None,
-                    autoclose=True,
-                    autoping=True,
-                    heartbeat=None,
-                    auth=None,
-                    origin=None,
-                    headers=None,
-                    proxy=None,
-                    proxy_auth=None,
-                    verify_ssl=None,
-                    fingerprint=None,
-                    ssl_context=None,
-                    proxy_headers=None,
-                    compress=0):
+    async def _ws_connect(self, url, *,
+                          protocols=(),
+                          timeout=10.0,
+                          receive_timeout=None,
+                          autoclose=True,
+                          autoping=True,
+                          heartbeat=None,
+                          auth=None,
+                          origin=None,
+                          headers=None,
+                          proxy=None,
+                          proxy_auth=None,
+                          ssl=None,
+                          verify_ssl=None,
+                          fingerprint=None,
+                          ssl_context=None,
+                          proxy_headers=None,
+                          compress=0):
 
         if headers is None:
             headers = CIMultiDict()
@@ -443,16 +527,16 @@ class ClientSession:
             extstr = ws_ext_gen(compress=compress)
             headers[hdrs.SEC_WEBSOCKET_EXTENSIONS] = extstr
 
+        ssl = _merge_ssl_params(ssl, verify_ssl, ssl_context, fingerprint)
+
         # send request
-        resp = yield from self.get(url, headers=headers,
-                                   read_until_eof=False,
-                                   auth=auth,
-                                   proxy=proxy,
-                                   proxy_auth=proxy_auth,
-                                   verify_ssl=verify_ssl,
-                                   fingerprint=fingerprint,
-                                   ssl_context=ssl_context,
-                                   proxy_headers=proxy_headers)
+        resp = await self.get(url, headers=headers,
+                              read_until_eof=False,
+                              auth=auth,
+                              proxy=proxy,
+                              proxy_auth=proxy_auth,
+                              ssl=ssl,
+                              proxy_headers=proxy_headers)
 
         try:
             # check handshake
@@ -461,7 +545,7 @@ class ClientSession:
                     resp.request_info,
                     resp.history,
                     message='Invalid response status',
-                    code=resp.status,
+                    status=resp.status,
                     headers=resp.headers)
 
             if resp.headers.get(hdrs.UPGRADE, '').lower() != 'websocket':
@@ -469,7 +553,7 @@ class ClientSession:
                     resp.request_info,
                     resp.history,
                     message='Invalid upgrade header',
-                    code=resp.status,
+                    status=resp.status,
                     headers=resp.headers)
 
             if resp.headers.get(hdrs.CONNECTION, '').lower() != 'upgrade':
@@ -477,7 +561,7 @@ class ClientSession:
                     resp.request_info,
                     resp.history,
                     message='Invalid connection header',
-                    code=resp.status,
+                    status=resp.status,
                     headers=resp.headers)
 
             # key calculation
@@ -489,7 +573,7 @@ class ClientSession:
                     resp.request_info,
                     resp.history,
                     message='Invalid challenge response',
-                    code=resp.status,
+                    status=resp.status,
                     headers=resp.headers)
 
             # websocket protocol
@@ -516,21 +600,22 @@ class ClientSession:
                             resp.request_info,
                             resp.history,
                             message=exc.args[0],
-                            code=resp.status,
+                            status=resp.status,
                             headers=resp.headers)
                 else:
                     compress = 0
                     notakeover = False
 
             proto = resp.connection.protocol
+            transport = resp.connection.transport
             reader = FlowControlDataQueue(
                 proto, limit=2 ** 16, loop=self._loop)
             proto.set_parser(WebSocketReader(reader), reader)
-            resp.connection.writer.set_tcp_nodelay(True)
+            tcp_nodelay(transport, True)
             writer = WebSocketWriter(
-                resp.connection.writer, use_mask=True,
+                proto, transport, use_mask=True,
                 compress=compress, notakeover=notakeover)
-        except Exception:
+        except BaseException:
             resp.close()
             raise
         else:
@@ -612,7 +697,7 @@ class ClientSession:
             self._request(hdrs.METH_DELETE, url,
                           **kwargs))
 
-    def close(self):
+    async def close(self):
         """Close underlying connector.
 
         Release all acquired resources.
@@ -622,8 +707,6 @@ class ClientSession:
                 self._connector.close()
             self._connector = None
 
-        return deprecated_noop('ClientSession.close() is a coroutine')
-
     @property
     def closed(self):
         """Is client session closed.
@@ -660,79 +743,78 @@ class ClientSession:
         self._connector = None
 
     def __enter__(self):
-        warnings.warn("Use async with instead", DeprecationWarning)
-        return self
+        raise TypeError("Use async with instead")
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        self.close()
+        # __exit__ should exist in pair with __enter__ but never executed
+        pass  # pragma: no cover
 
-    if PY_35:
-        @asyncio.coroutine
-        def __aenter__(self):
-            return self
+    async def __aenter__(self):
+        return self
 
-        @asyncio.coroutine
-        def __aexit__(self, exc_type, exc_val, exc_tb):
-            yield from self.close()
+    async def __aexit__(self, exc_type, exc_val, exc_tb):
+        await self.close()
 
 
-class _BaseRequestContextManager(_BaseCoroMixin):
+class _BaseRequestContextManager(Coroutine):
 
-    __slots__ = ('_resp',)
+    __slots__ = ('_coro', '_resp')
 
     def __init__(self, coro):
-        super().__init__(coro)
         self._coro = coro
 
-    if PY_35:
-        @asyncio.coroutine
-        def __aenter__(self):
-            self._resp = yield from self._coro
-            return self._resp
+    def send(self, arg):
+        return self._coro.send(arg)
+
+    def throw(self, arg):
+        return self._coro.throw(arg)
+
+    def close(self):
+        return self._coro.close()
+
+    def __await__(self):
+        ret = self._coro.__await__()
+        return ret
+
+    def __iter__(self):
+        return self.__await__()
+
+    async def __aenter__(self):
+        self._resp = await self._coro
+        return self._resp
 
 
 class _RequestContextManager(_BaseRequestContextManager):
-    if PY_35:
-        @asyncio.coroutine
-        def __aexit__(self, exc_type, exc, tb):
-            # We're basing behavior on the exception as it can be caused by
-            # user code unrelated to the status of the connection.  If you
-            # would like to close a connection you must do that
-            # explicitly.  Otherwise connection error handling should kick in
-            # and close/recycle the connection as required.
-            self._resp.release()
+    async def __aexit__(self, exc_type, exc, tb):
+        # We're basing behavior on the exception as it can be caused by
+        # user code unrelated to the status of the connection.  If you
+        # would like to close a connection you must do that
+        # explicitly.  Otherwise connection error handling should kick in
+        # and close/recycle the connection as required.
+        self._resp.release()
 
 
 class _WSRequestContextManager(_BaseRequestContextManager):
-    if PY_35:
-        @asyncio.coroutine
-        def __aexit__(self, exc_type, exc, tb):
-            yield from self._resp.close()
+    async def __aexit__(self, exc_type, exc, tb):
+        await self._resp.close()
 
 
-class _SessionRequestContextManager(_RequestContextManager):
+class _SessionRequestContextManager:
 
-    __slots__ = _RequestContextManager.__slots__ + ('_session', )
+    __slots__ = ('_coro', '_resp', '_session')
 
     def __init__(self, coro, session):
-        super().__init__(coro)
+        self._coro = coro
+        self._resp = None
         self._session = session
 
-    @asyncio.coroutine
-    def __iter__(self):
-        try:
-            return (yield from self._coro)
-        except BaseException:
-            yield from self._session.close()
-            raise
+    async def __aenter__(self):
+        self._resp = await self._coro
+        return self._resp
 
-    if PY_35:
-        def __await__(self):
-            try:
-                return (yield from self._coro)
-            except BaseException:
-                yield from self._session.close()
-                raise
+    async def __aexit__(self, exc_type, exc_val, exc_tb):
+        self._resp.close()
+        await self._session.close()
 
 
 def request(method, url, *,
@@ -745,7 +827,6 @@ def request(method, url, *,
             auth=None,
             allow_redirects=True,
             max_redirects=10,
-            encoding=None,
             version=http.HttpVersion11,
             compress=None,
             chunked=None,
@@ -782,10 +863,10 @@ def request(method, url, *,
     loop - Optional event loop.
     Usage::
       >>> import aiohttp
-      >>> resp = yield from aiohttp.request('GET', 'http://python.org/')
+      >>> resp = await aiohttp.request('GET', 'http://python.org/')
       >>> resp
       <ClientResponse(python.org/) [200]>
-      >>> data = yield from resp.read()
+      >>> data = await resp.read()
     """
     connector_owner = False
     if connector is None:
@@ -806,11 +887,10 @@ def request(method, url, *,
                          auth=auth,
                          allow_redirects=allow_redirects,
                          max_redirects=max_redirects,
-                         encoding=encoding,
                          compress=compress,
                          chunked=chunked,
                          expect100=expect100,
                          read_until_eof=read_until_eof,
                          proxy=proxy,
                          proxy_auth=proxy_auth,),
-        session=session)
+        session)
