@@ -453,7 +453,7 @@ class BodyPartReaderPayload(Payload):
         field = self._value
         chunk = await field.read_chunk(size=2**16)
         while chunk:
-            writer.write(field.decode(chunk))
+            await writer.write(field.decode(chunk))
             chunk = await field.read_chunk(size=2**16)
 
 
