@@ -256,7 +256,7 @@ class RequestHandler(asyncio.streams.FlowControlMixin, asyncio.Protocol):
                             # don't set result twice
                             waiter.set_result(None)
 
-                self._upgraded = upgraded
+                self._upgrade = upgraded
                 if upgraded and tail:
                     self._message_tail = tail
 
